#include frex:shaders/api/accessibility.glsl
#include frex:shaders/api/fog.glsl
#include frex:shaders/api/light.glsl
#include frex:shaders/api/material.glsl
#include frex:shaders/api/player.glsl
#include frex:shaders/api/view.glsl
#include frex:shaders/api/world.glsl
#include frex:shaders/lib/math.glsl
#include ecos:shaders/lib/tonemap.glsl

#define GODRAYS

uniform sampler2D glint_item;
uniform sampler2D glint_entity;
uniform sampler2D colored_lights;

in float v_diffuse;
in float v_skyDiffuse;
in float v_heldCosIn;
in float v_heldCosOut;

#ifdef SHADOW_MAP_PRESENT
in vec4 v_shadowViewPos;
#endif

out vec4 fragColor;
out vec4 fragBloom;

#ifdef SHADOW_MAP_PRESENT
// Helper function
vec3 shadowDist(int cascade)
{
  vec4 c = frx_shadowCenter(cascade);
  return abs((c.xyz - v_shadowViewPos.xyz) / c.w);
}

// Function for obtaining the cascade level
int selectShadowCascade()
{
  vec3 d3 = shadowDist(3);
  vec3 d2 = shadowDist(2);
  vec3 d1 = shadowDist(1);

  int cascade = 0;

  if (d3.x < 1.0 && d3.y < 1.0 && d3.z < 1.0) {
    cascade = 3;
  } else if (d2.x < 1.0 && d2.y < 1.0 && d2.z < 1.0) {
    cascade = 2;
  } else if (d1.x < 1.0 && d1.y < 1.0 && d1.z < 1.0) {
    cascade = 1;
  }

  return cascade;
}
#endif

vec3 getGlint()
{
  // vanilla scale factor for entity, works in most scenario
  const float scale = 0.16;

  // vanilla translation factor
  float time = frx_renderSeconds * 8. * frx_glintSpeed;
  vec2 translation = vec2(-time / 110., time / 30.);
  vec2 uv = frx_normalizeMappedUV(frx_texcoord) * scale + translation;
  vec3 glint = mix(texture(glint_item, uv).rgb, texture(glint_entity, uv).rgb, float(frx_matGlintEntity));

  // emulate GL_SRC_COLOR sfactor
  return glint * glint * frx_glintStrength * float(frx_matGlint);
}

float shiny_glint_noise(vec2 uv, float zoom)
{
    const float skew = 0.4;
    const float detail = 8.0;
    uv += uv.y * skew;
    uv *= zoom;
    float t = mod(uv.x, 1.0);
    t *= detail;
    float f = fract(t);
    t -= f;
    t /= detail;
    f -= 0.5;
    f = abs(f);
    f = 1.0 - 2.0 * f;
    float n = frx_noise2d(vec2(t)) * f;
    return n * n;
    // uv *= 0.0001;
    // uv = mod(uv, 1.0);
    // float n = frx_noise2d(uv.xy);
    // return n;
}

vec3 getHeldLight() {
	vec4 held = frx_heldLight;

  if (frx_heldLight.w <= 0) {
    return vec3(0.0);
  }

  float d = min(frx_distance / (held.w * 14.0), 1.0);
  d = 1.0 - d;

  float angle = dot(frx_cameraView, frx_vertex.xyz / frx_distance);
  angle = smoothstep(v_heldCosOut, v_heldCosIn, angle);

  return frx_heldLight.rgb * angle * d;
}

vec3 getLight() {
    float fake_diffuse = frx_fragEnableDiffuse ? v_diffuse : 1.0;

    if (frx_isGui && !frx_isHand) {
      return vec3(fake_diffuse);
    }

  #ifdef SHADOW_MAP_PRESENT
    // Obtain the cascade level
    int cascade = selectShadowCascade();

    // Obtain shadow-space position
    vec4 shadowPos = frx_shadowProjectionMatrix(cascade) * v_shadowViewPos;

    // Transform into texture coordinates
    vec3 shadowTexCoord = shadowPos.xyz * 0.5 + 0.5;

    // Sample the shadow map
    float directSkyLight = texture(frxs_shadowMap, vec4(shadowTexCoord.xy, float(cascade), shadowTexCoord.z));
  #else
    float directSkyLight = 1.0;
  #endif    

    // Multiply the color once more with the diffuse shading
    if (frx_fragEnableDiffuse) {
      directSkyLight = min(directSkyLight, v_skyDiffuse);
    }

    float skyStrength = mix(2.0, 0.4, frx_skyLightTransitionFactor * float(frx_worldIsMoonlit)) * (1.0 - 0.5 * frx_rainGradient - 0.25 * frx_thunderGradient);
    directSkyLight *= skyStrength;

    // Apply transition factor for sunrise and sunset
    directSkyLight *= frx_skyLightTransitionFactor;
  
    vec3 block_light;
    vec3 sun_light = vec3(directSkyLight);
    vec3 light_color = vec3(0.0);
    vec3 ambient_sky_light = vec3(0.5) * skyStrength;

    vec3 night_ambient = mix(vec3(1.0), vec3(0.3, 0.3, 1.0), frx_skyLightTransitionFactor * float(frx_worldIsMoonlit));
    sun_light *= night_ambient;
    ambient_sky_light *= night_ambient;

    // Always wrap vanilla lighting operation within #ifdef VANILLA_LIGHTING directive
    #ifdef VANILLA_LIGHTING
      #ifndef COLORED_LIGHTS_ENABLED
        block_light = vec3(frx_fragLight.x);
      #endif
      sun_light *= frx_fragLight.y;
      ambient_sky_light *= frx_fragLight.y;
    #endif

    #ifdef COLORED_LIGHTS_ENABLED
      block_light = frx_getLight(colored_lights, (frx_vertex.xyz + frx_vertexNormal * 0.1) * (1.0 - float(frx_isHand)) + frx_cameraPos, block_light);
    #endif

    block_light = pow(block_light, vec3(5.0)) * 2.0 * fake_diffuse;

    vec3 ambient_light = vec3(0.1 + (1.0 - float(frx_worldHasSkylight)) * 0.3);
    light_color += block_light + sun_light + ambient_sky_light + ambient_light + getHeldLight();

    // Multiply it by ambient occlusion factor
    if (frx_fragEnableAo) {
      light_color *= frx_fragLight.z * frx_fragLight.z;
    }

    // Apply emissivity (yes we apply it at bloom as well)
    light_color += vec3(frx_fragEmissive);

    // Soften strong lights
    light_color = sqrt(light_color * (0.25 + 0.25 * frx_viewBrightness));

    return light_color;
}

void frx_pipelineFragment()
{
  // Obtain true color by multiplying sprite color (usually texture) and vertex color (usually biome color)
  vec4 color = frx_fragColor;

  // Finally, multiply the fragment color by the light color
  vec3 light = getLight();
  color.rgb *= light;

  // Apply glint
  color.rgb += getGlint();
  
  float bloom = 0.0;

  if (!frx_isGui || frx_isHand) {
    // Apply flash and hurt effect
    color.rgb += float(frx_matFlash);
    color.rgb += vec3(0.25 * float(frx_matHurt), 0.0, 0.0);

    #ifdef PBR_ENABLED
      mat3 TBN = mat3(frx_vertexTangent.xyz, cross(frx_vertexTangent.xyz, frx_vertexNormal) * frx_vertexTangent.w, frx_vertexNormal);
      frx_fragNormal = TBN * frx_fragNormal;
      vec3 fragView = -frx_vertex.xyz / frx_distance;
      vec3 reflection = reflect(fragView, frx_fragNormal);
      reflection = TBN * reflection;
      reflection *= sign(reflection.z);
      vec2 shiny_glint_uv = reflection.xy / (reflection.z + 0.000001);
      float alpha = 1.0 - frx_fragRoughness;
      float alpha2 = alpha * alpha;
      float shiny_glint = shiny_glint_noise(shiny_glint_uv, 0.4 * alpha2) * alpha2 * reflection.z;
      color.rgb += mix(0.5 * light * shiny_glint, (shiny_glint - 0.5 * alpha) * light * frx_fragColor.rgb, alpha * frx_fragReflectance);
      color.a += (1.0 - color.a) * pow(max(0.0, 1.0 - dot(fragView, frx_vertexNormal)), 10.0);
    #endif
    
    // Tonemap for funsies (see footnote 1)
    color.rgb = tonemap(color.rgb);

    // Get bloom amount (see footnote 1)
    bloom = getBloom(color.rgb, frx_fragEmissive);
  
    // Fog
    float fogDensity = pow(smoothstep(0.0, 512.0, frx_distance), 0.4) * pow(smoothstep(512.0, -32.0, frx_vertex.y), 4.0);
    float fogFactor = max(fogDensity, smoothstep(frx_fogStart, frx_fogEnd, frx_distance));
    color.rgb = mix(color.rgb, frx_fogColor.rgb, fogFactor);
    bloom = mix(bloom, 0, fogFactor);
  }

  // Write color data to the color attachment
  fragColor = color;// vec4(gl_FragCoord.x / frx_viewWidth, gl_FragCoord.y / frx_viewHeight, gl_FragCoord.z * gl_FragCoord.w, 1.0);
  
  // Write bloom data to the bloom attachment
  fragBloom = vec4(bloom * color.a, 0.0, 0.0, color.a);

  // Write position data to the depth attachment
  gl_FragDepth = gl_FragCoord.z;
}

// footnote 1: based on this calculation https://www.desmos.com/calculator/kmolpqoaxl